#include <stdio.h>
#include <stdlib.h>
#include "mpi.h"
#include <math.h>

#define xprocessors 2    /* # of Processors in x direcction. */
#define yprocessors 2    /* # of Processors in y direcction. */

#define NDIM 2           /* Number of Dimensions */ 


double fun(double x, double y);

int main(int argc, char *argv[])
{
    MPI_Status status;
    int         n = 1024;
    int         rank, numprocs, i,j;
    double      mypisquared, pisqured;
    int         coords[NDIM];
    int         num_proc_x, num_proc_y;
    double      x, y, sx, ex, sy, ey, h;
    double      pisquare = 0.0;
    int         reorder;
    int         dims[2], periods[2];
  
    MPI_Comm comm2d;

    MPI_Init(&argc, &argv);

    MPI_Comm_size(MPI_COMM_WORLD, &numprocs);
  

    periods[0] = 0;
    periods[1] = 0; 

    dims[0] = yprocessors;
    dims[1] = xprocessors;

    reorder = 1;       

/* The following functions are used to create the two dimentional grid
   and it is divided into the processors.   */


    MPI_Cart_create(MPI_COMM_WORLD, 2, dims, periods, reorder, &comm2d);

    MPI_Comm_rank(comm2d, &rank);

    MPI_Cart_coords(comm2d, rank, 2, coords);
    num_proc_y = coords[0];
    num_proc_x = coords[1];
       
        h = 1.0 /(double) n;

 /*    The following calculations are done by each process   
       individually. Since we have created a grid each process will     
       perform the operation in its own assigned area only.   */


       sx = (double)num_proc_x / (double)(2);
       ex = ((double)num_proc_x / (double)(2)) + 0.5;

       sy = (double)num_proc_y / (double)(2);
       ey = ((double)num_proc_y / (double)(2)) + 0.5;

       printf("I am processor #  %d\n", rank);
       printf("My x and y coordinates are\n");
       printf("From     %f , %f\n", sx,ex);  
       printf("To       %f , %f\n\n", sy,ey);  
       

	for (x=sx; x<=ex; x+=h)
	{
            for (y=sy; y<=ey; y+=h)    
            {
               pisquare += (h*h) * fun(x + 0.5*h, y + 0.5*h);       
/*	      pisquare += (h*h) *(fun(x,y) + fun(x,y+(0.5 * h)) + fun(x+(0.5 * h), y) + 
				  fun(x+(0.5 * h), y+(0.5 * h)))/ 4;
	      pisquare += h*h*(1/((1+(x+.5*h)*(x+.5*h))*(1+(y+.5*h)*(y+.5*h))));
*/
	      
            }
        }

        mypisquared = 16.0 * pisquare;
  
        MPI_Reduce(&mypisquared, &pisqured, 1, MPI_DOUBLE, MPI_SUM, 0, 
                    comm2d);
     
        if (rank == 0)
        {
       	   printf("The value of pisquared is approximately %.16f\n", 
           pisqured);
        }
 
  MPI_Finalize;

} /* End of main function */    



/* The following function calculates the value of integration. */

double fun(double x, double y)
{
    double value;
    value = 1/((1 + x*x) * (1 + y*y));
    return value;
}    /* end of fun function.  */
    

